import functools
import os
import logging
import pathlib
import shutil
from io import BytesIO

from typing import Optional, Callable

from pytube.exceptions import PytubeError
from pytube import YouTube, StreamQuery, Stream
from pytube.exceptions import RegexMatchError, VideoUnavailable

from suskabot.utils.utils import pytube_video_resolution_to_int, bytes_to_megabytes


logger: logging.Logger = logging.getLogger(__name__)


class VideoTooLarge(Exception):
    """Raise if a video file size exceeds set limits"""


class YTVideoMetadata:
    title: str
    filesize_bytes: int

    def __init__(
            self,
            title: str,
            filesize_bytes: int,
    ) -> None:
        self.title = title
        self.filesize_bytes = filesize_bytes


class YTDownloaderService:
    _MINIMUM_VIDEO_RESOLUTION: int
    _MAXIMUM_VIDEO_RESOLUTION: int
    _FILE_SIZE_LIMIT_MB: int
    _SAVE_TO_DRIVE: bool

    _DEFAULT_DOWNLOAD_PATH = pathlib.Path(os.getcwd()).joinpath("yt_downloads/")

    def __init__(
            self,
            min_video_res: int,
            max_video_res: int,
            file_size_limit_mb: int,
            save_to_drive: bool
    ) -> None:
        self._MINIMUM_VIDEO_RESOLUTION = min_video_res
        self._MAXIMUM_VIDEO_RESOLUTION = max_video_res
        self._FILE_SIZE_LIMIT_MB = file_size_limit_mb
        self._SAVE_TO_DRIVE = save_to_drive

    def __get_best_quality_stream_under_limits(self, video_streams: StreamQuery) -> Optional[Stream]:
        ordered_video_streams: StreamQuery = video_streams.order_by("resolution").desc()

        video_stream: Stream
        for video_stream in ordered_video_streams:
            video_resolution: int = pytube_video_resolution_to_int(video_stream.resolution)
            video_filesize_mb: int = bytes_to_megabytes(video_stream.filesize)

            # noinspection PyChainedComparisons
            # I don't like how chained comparisons look.
            if (video_filesize_mb <= self._FILE_SIZE_LIMIT_MB and
                    video_resolution >= self._MINIMUM_VIDEO_RESOLUTION and
                    video_resolution <= self._MAXIMUM_VIDEO_RESOLUTION):
                return video_stream
        else:
            return None

    # all the stream availability exceptions are raised here!
    def get_youtube_video_stream_metadata(self, video_url: str) -> Optional[YTVideoMetadata]:
        try:
            video_data = YouTube(url=video_url)
            video_data.check_availability()
        except RegexMatchError as e:
            raise e
        except VideoUnavailable as e:
            raise e

        progressive_video_streams: StreamQuery = video_data.streams.filter(progressive=True)

        video_stream: Stream | None = self.__get_best_quality_stream_under_limits(progressive_video_streams)
        if not video_stream:
            raise VideoTooLarge(f"Can't download videos larger than {self._FILE_SIZE_LIMIT_MB}MB")

        metadata = YTVideoMetadata(
            title=video_data.title,
            filesize_bytes=video_stream.filesize
        )

        return metadata

    # here it is assumed that the best video stream under limits is ready to be downloaded
    def download_youtube_video(
            self,
            video_url: str,
            progress_callback: Callable[[Stream, bytes, int], None] | None = None
    ) -> bytes | None:
        if progress_callback:
            video_data = YouTube(
                url=video_url,
                on_progress_callback=progress_callback
            )
        else:
            video_data = YouTube(url=video_url)

        progressive_video_streams: StreamQuery = video_data.streams.filter(progressive=True)

        video_stream: Stream = self.__get_best_quality_stream_under_limits(progressive_video_streams)

        if self._SAVE_TO_DRIVE:
            video_path: str = video_stream.download(
                output_path=str(self._DEFAULT_DOWNLOAD_PATH)
            )
        else:
            with BytesIO() as buffer:
                video_stream.stream_to_buffer(buffer)
                logger.info(f"Successfully downloaded {video_url} to RAM")
                return buffer.getvalue()

        with open(video_path, "rb") as video_file:
            logger.info(f"Successfully downloaded {video_url} to drive")
            return video_file.read()

    # doesn't serve any purpose as of now, but I'll leave it here for a bit
    def __clear_download_directory(self) -> None:
        try:
            shutil.rmtree(self._DEFAULT_DOWNLOAD_PATH)
        except Exception as e:
            logger.warning(f"Could not clear download directory, {e}")
            raise e



import logging
import sys
import telegram.ext

from typing import Optional

from config import config

from suskabot.controllers.telegram_api import bot
from suskabot.services import translator
from suskabot.services import youtube_downloader


logging.basicConfig(
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
    level=logging.INFO,
    force=True
)
if logging.root.level > logging.DEBUG:
    # prevents log spam from infinite polling
    logging.getLogger("httpx").setLevel(logging.WARNING)


logger = logging.getLogger(__name__)


if __name__ == '__main__':
    try:
        config: Optional[config.Config] = config.load_config()
    except AttributeError:
        # somehow pydantic raises an attribute error when trying to access config.Config for the type hint
        # removing the type hint solves the issue, but I'd rather just suppress the pydantic error
        pass
    if not config:
        logger.error("Couldn't load config file")
        sys.exit(1)

    # services
    translation_service = translator.TranslatorService(
        comma_separated_translation_services=config.comma_separated_translation_services,
        user_language=config.user_language,
        default_lang_to_translate_to=config.default_language_to_translate_to
    )
    logger.debug("Successfully instantiated translation service")

    yt_downloader_service = youtube_downloader.YTDownloaderService(
        file_size_limit_mb=config.file_size_limit_mb,
        min_video_res=config.minimum_video_resolution,
        max_video_res=config.maximum_video_resolution,
        save_to_drive=config.save_to_drive
    )
    logger.debug("Successfully instantiated youtube downloader service")

    # controllers
    app: telegram.ext.Application = (telegram.ext.ApplicationBuilder()
                                     .token(config.tg_bot_api_token)
                                     .concurrent_updates(True)
                                     .media_write_timeout(60)
                                     .build()
                                     )
    logger.debug("Successfully instantiated python-telegram-bot application")

    tg_bot = bot.TGBot(
        application=app,
        translator_service=translation_service,
        yt_downloader_service=yt_downloader_service
    )
    logger.debug("Successfully instantiated suskabot application")

    tg_bot.start()



import gc
import unittest

from pytube.exceptions import RegexMatchError

from suskabot.services.youtube_downloader import YTDownloaderService


class TestYoutubeDownloader(unittest.TestCase):
    min_video_resolution: int = 144
    max_video_resolution: int = 2160
    file_size_limit_mb: int = 10
    save_to_drive: bool = False

    test_video_url: str = "https://www.youtube.com/watch?v=BBJa32lCaaY"
    test_video: bytes

    service: YTDownloaderService

    def setUp(self):
        self.service = YTDownloaderService(
            file_size_limit_mb=self.file_size_limit_mb,
            min_video_res=self.min_video_resolution,
            max_video_res=self.max_video_resolution,
            save_to_drive=self.save_to_drive
        )

        with open("./suskabot/tests/resources/cats.mp4", "rb") as test_file:
            self.test_video = test_file.read()

    def test_correct_downloads(self):
        self.assertEqual(
            self.service.download_youtube_video(self.test_video_url),
            self.test_video
        )

    def test_invalid_url_detection(self):
        self.assertRaises(RegexMatchError, self.service.download_youtube_video, "hehe")

    def tearDown(self):
        del self.service
        gc.collect()


if __name__ == '__main__':
    unittest.main()




def get_formatted_progressbar(video_title: str, progress_percentage: int) -> str:
    blank_symbol = "▯ "
    filled_symbol = "▮ "
    total_symbols_number = 10

    filled_symbols_number = progress_percentage // total_symbols_number
    blank_symbols_number = total_symbols_number - filled_symbols_number

    progressbar: str = \
        (f"Downloading {video_title}..\n"
         f"{filled_symbol * filled_symbols_number}{blank_symbol * blank_symbols_number}{progress_percentage}%")

    return progressbar


def get_download_progress_percentage(total_bytes: int, remaining_bytes: int) -> int:
    return int(((total_bytes - remaining_bytes) / total_bytes) * 100)


def pytube_video_resolution_to_int(resolution: str) -> int:
    return int(resolution[0:-1])


def bytes_to_megabytes(bytes_number: int) -> int:
    return int(bytes_number / 1024 / 1024)



import logging
import os
from typing import Self, Dict, Optional

import telegram.constants
import pydantic
import dotenv


logger = logging.getLogger(__name__)


POSSIBLE_VIDEO_RESOLUTIONS = [
    144,
    240,
    360,
    480,
    720,
    1080,
    1440,
    2160
]


class Config(pydantic.BaseModel):
    tg_bot_api_token: str = pydantic.Field(pattern=r"^[0-9]{8,10}:[a-zA-Z0-9_-]{35}$")

    comma_separated_translation_services: str = pydantic.Field(
        default="google,yandex,bing,argos",
        pattern=r"^([a-zA-Z]+,?)+"
    )
    user_language: str = pydantic.Field(
        default="ru",
        pattern=r"^[^-]{2,3}(-[^-]{2,3})?(-[^-]{2,3})?$"
    )
    default_language_to_translate_to: str = pydantic.Field(
        default="en",
        pattern=r"^[^-]{2,3}(-[^-]{2,3})?(-[^-]{2,3})?$",
    )

    # right now only progressive videos are supported, so video quality is 720p max.
    # it also looks like 360p and 720p are the only available options
    minimum_video_resolution: int = POSSIBLE_VIDEO_RESOLUTIONS[0]
    maximum_video_resolution: int = POSSIBLE_VIDEO_RESOLUTIONS[-1]

    @pydantic.field_validator("maximum_video_resolution", "minimum_video_resolution")
    def validate_option(cls, v):
        assert v in POSSIBLE_VIDEO_RESOLUTIONS, ("this does not look like a valid resolution. "
                                                 f"supported resolutions: {POSSIBLE_VIDEO_RESOLUTIONS}")
        return v

    @pydantic.model_validator(mode='after')
    def check_if_valid_video_resolution_preferences(self) -> Self:
        if self.minimum_video_resolution > self.maximum_video_resolution:
            raise ValueError("minimum video resolution can't be higher than maximum video resolution")
        if self.minimum_video_resolution > 720:
            logger.warning("Since only progressive videos are supported right now,"
                           " 720p is the maximum possible resolution")
        return self

    # setting the limit higher than the default will break telegram api controller!
    file_size_limit_mb: int = pydantic.Field(
        default=telegram.constants.FileSizeLimit.FILESIZE_UPLOAD.value,
        gt=0,
        validate_default=True
    )

    save_to_drive: bool = pydantic.Field(
        default=True
    )

    # TODO: having a config for yt video downloads directory seems like a nice idea


def load_config() -> Optional[Config]:
    if not os.path.exists("./suskabot/config/.env.secret"):
        with open("./suskabot/config/.env.secret", "w") as f:
            f.write("TG_BOT_API_TOKEN=")
        logger.info("Created .env.secret file")

    public_env_file_config: Dict[str, str | None] = dotenv.dotenv_values("./suskabot/config/.env.public")

    secret_env_file_config: Dict[str, str | None] = dotenv.dotenv_values("./suskabot/config/.env.secret")

    config_data: Dict[str, str | None] = {}
    # it is assumed that every valid configuration is type annotated
    # I tried really hard, but there's no easier way to achieve the same functionality without repeating config names
    configurations: set[str] = vars(Config)['__annotations__'].keys()
    configuration: str
    for configuration in configurations:
        configuration_value: str | None
        configuration_upper = configuration.upper()
        logger.debug(f"Getting {configuration}..")
        if configuration_upper in os.environ:
            configuration_value = os.environ[configuration_upper]
            logger.debug(f"Got {configuration} from environ")
        elif configuration_upper in secret_env_file_config:
            configuration_value = secret_env_file_config[configuration_upper]
            logger.debug(f"Got {configuration} from env.secret")
        elif configuration_upper in public_env_file_config:
            configuration_value = public_env_file_config[configuration_upper]
            logger.debug(f"Got {configuration} from env.public")
        else:
            logger.debug(f"Couldn't find {configuration} anywhere")
            continue

        config_data[configuration] = configuration_value
        logger.info(f"{configuration} set to {configuration_value}")

    config: Config = Config(**config_data)

    return config



import gc
import unittest

from suskabot.services.translator import TranslatorService


class TestTranslator(unittest.TestCase):
    translation_services: str = "google,yandex,bing,argos"
    mother_tongue: str = "ru"
    default_lang_to_translate_to: str = "en"

    service: TranslatorService

    def setUp(self):
        self.service = TranslatorService(
            comma_separated_translation_services=self.translation_services,
            user_language=self.mother_tongue,
            default_lang_to_translate_to=self.default_lang_to_translate_to
        )

    def test_number_translation(self):
        self.assertEqual(self.service.translate("пять"), "five")

    def test_reverse_language_translation(self):
        self.assertEqual(self.service.translate("five"), "пять")

    def test_well_known_translation(self):
        self.assertEqual(self.service.translate("кошка"), "cat")

    def tearDown(self):
        del self.service
        gc.collect()


if __name__ == '__main__':
    unittest.main()



import asyncio
import logging
import concurrent.futures
from asyncio import AbstractEventLoop
from typing import  Set

import telegram.error
from pytube import Stream
from pytube.exceptions import RegexMatchError, VideoUnavailable, PytubeError
from telegram import Update, Message
from telegram.ext import ContextTypes, Application, MessageHandler, filters

from suskabot.services.youtube_downloader import YTDownloaderService, VideoTooLarge, YTVideoMetadata

from suskabot.utils.utils import get_download_progress_percentage, get_formatted_progressbar

logger: logging.Logger = logging.getLogger(__name__)


class YTDownloaderHandlers:
    _application: Application
    _yt_downloader_service: YTDownloaderService

    # set of telegram user ids
    _users_with_a_download_in_progress: Set[int] = set()

    _asyncio_loop: AbstractEventLoop = asyncio.get_event_loop()

    def __init__(
            self,
            application: Application,
            downloader_service: YTDownloaderService
    ) -> None:
        self._application = application
        self._yt_downloader_service = downloader_service

        self._application.add_handlers(
            [
                MessageHandler(
                    filters.Regex(r"^(https?\:\/\/)?((www\.)?youtube\.com|youtu\.be)\/.+$"), self.send_yt_video
                ),
            ]
        )

    async def send_yt_video(self, update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
        if update.effective_user.id in self._users_with_a_download_in_progress:
            logger.info(f"{update.effective_user.full_name}, {update.effective_user.id} "
                        "tried to download another video before his previous download finished")
            await update.effective_message.reply_text("Wait for your previous download to finish!")
            return

        video_url: str = update.effective_message.text

        try:
            video_metadata: YTVideoMetadata = self._yt_downloader_service.get_youtube_video_stream_metadata(video_url)
        except RegexMatchError:
            logger.error(f"{video_url} doesnt look like a valid url, sussy baka: {update.effective_user.full_name}")
            await update.effective_message.reply_text("Check video url")
            return
        except VideoUnavailable as e:
            logger.error(f"{video_url} is not available, {e}")
            await update.effective_message.reply_text(f"The video is not available, {e}")
            return
        except VideoTooLarge as e:
            logger.error(f"Video too larger error, {e}")
            await update.effective_message.reply_text(f"Video too larger error, {e}")
            return

        progressbar: str = get_formatted_progressbar(video_metadata.title, 0)
        progressbar_message: Message = await update.effective_message.reply_text(progressbar)

        self._users_with_a_download_in_progress.add(update.effective_user.id)
        logger.debug(f"Added {update.effective_user.id} to the set of downloading user, "
                     f"current set:{self._users_with_a_download_in_progress}")

        async def _clean_up():
            await progressbar_message.delete()
            self._users_with_a_download_in_progress.remove(update.effective_user.id)
            logger.debug(f"Successfully deleted the progressbar and "
                         f"Removed {update.effective_user.id} from the set of downloading user, "
                         f"current set:{self._users_with_a_download_in_progress}")

        def download_progress_callback(stream: Stream, b: bytes, remaining_bytes: int) -> None:
            logger.debug(f"Got a progress callback for {update.effective_user.full_name} {update.effective_user.id}, "
                         f"{remaining_bytes=}")
            del stream, b  # unused for current implementation of the callback

            progress_percentage: int = get_download_progress_percentage(
                video_metadata.filesize_bytes,
                remaining_bytes
            )

            updated_progressbar: str = get_formatted_progressbar(
                video_metadata.title,
                progress_percentage
            )

            asyncio.run_coroutine_threadsafe(
                context.bot.edit_message_text(
                    text=updated_progressbar,
                    chat_id=update.effective_chat.id,
                    message_id=progressbar_message.id
                ),
                self._asyncio_loop
            )

        download_yt_video = self._yt_downloader_service.download_youtube_video
        this_loop = self._asyncio_loop
        with concurrent.futures.ThreadPoolExecutor() as pool:
            logger.debug(f"Spinning up the youtube video downloader thread for "
                         f"{update.effective_user.full_name} {update.effective_user.id}")
            try:
                downloaded_video: bytes | None = await this_loop.run_in_executor(
                    pool,
                    download_yt_video,
                    video_url,
                    download_progress_callback
                )
            except (OSError, PytubeError) as e:
                await _clean_up()
                logger.error(f"Couldn't download video to disk, {e}")
                await update.effective_message.reply_text("Service error, please try again later")
                return

        await progressbar_message.edit_text("Download complete! Sending the video..")
        logger.debug("Video downloaded, starting to send it")

        try:
            await context.bot.send_video(
                chat_id=update.effective_chat.id,
                video=downloaded_video,
                caption=video_metadata.title
            )
        except telegram.error.TelegramError as e:
            logger.error(f"Couldn't send downloaded video, {e}")
            await update.effective_message.reply_text(
                "A Telegram error occurred"
            )
        finally:
            await _clean_up()



import logging

from telegram import Update
from telegram.ext import ContextTypes, Application, CommandHandler


logger: logging.Logger = logging.getLogger(__name__)


class AppHandlers:
    _application: Application

    def __init__(self, application: Application) -> None:
        self._application = application

        self._application.add_handlers(
            (
                CommandHandler('start', self.start),
            )
        )

    async def start(self, update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
        greeting_text = "Hi! I'm a Telegram bot that was made to make my creators life easier, " \
                        "here's some of my functionality:\n" \
                        "1) Translator\n" \
                        "English messages will be translated to Russian\n" \
                        "Messages on the any other language will be translated to English\n\n" \
                        "2) YT videos downloader\n" \
                        "Provide a url link to a YouTube video to get mp4 video of it!\n\n" \
                        "3) TBD"

        await context.bot.send_message(
            chat_id=update.effective_chat.id,
            text=greeting_text
        )

        logger.debug(f"Sent the greeting text to {update.effective_user.full_name}, {update.effective_user.id}")



import logging
from typing import List, Dict

import translators

from lingua import Language, LanguageDetector, LanguageDetectorBuilder
from translators.server import TranslatorError

logger = logging.getLogger(__name__)


class TranslatorService:
    _translation_services: List[str]
    _user_language: str
    _default_language_to_translate_to: str

    def __init__(
            self,
            comma_separated_translation_services: str,
            user_language: str,
            default_lang_to_translate_to: str
    ) -> None:
        translation_services: List[str] = comma_separated_translation_services.strip().split(',')
        self._translation_services = translation_services

        self._user_language = user_language
        self._default_language_to_translate_to = default_lang_to_translate_to

    def __get_language_to_translate_to(self, language_code: str) -> str:
        # detected language: language to translate to
        translation_rules: Dict[str, str] = {
            self._default_language_to_translate_to: self._user_language,
        }
        default_language_to_translate_to: str = self._default_language_to_translate_to

        language_to_translate_to: str = translation_rules.get(language_code, default_language_to_translate_to)

        logger.debug(f"Will translate from {language_code} to {language_to_translate_to}")
        return language_to_translate_to

    def translate(self, text_to_translate: str) -> str:
        supported_languages: List[Language] = [Language.ENGLISH, Language.RUSSIAN, Language.UKRAINIAN]

        language_detector: LanguageDetector = LanguageDetectorBuilder.from_languages(*supported_languages).build()
        language_to_translate_from: str = language_detector.detect_language_of(text_to_translate).name.lower()[0:2]
        logger.debug(f"Detected {language_to_translate_from} language from \"{text_to_translate}\"")

        translation_services: List[str] = self._translation_services

        language_to_translate_to: str = self.__get_language_to_translate_to(language_to_translate_from)

        logger.debug(f"Will use this list of translation services: {translation_services}")
        for translation_service in translation_services:
            logger.debug(f"Trying to translate using {translation_service}..")
            try:
                translation: str = translators.translate_text(
                    query_text=text_to_translate,
                    translator=translation_service,
                    from_language=language_to_translate_from,
                    to_language=language_to_translate_to,
                    if_use_preacceleration=False
                )
            except TranslatorError as e:
                logger.warning(f"{e} occurred when trying to translate \"{text_to_translate}\" "
                               f"from \"{language_to_translate_from}\" "
                               f"to \"{language_to_translate_to}\" "
                               f"using {translation_service} service")
                continue
            except Exception as e:
                logger.warning(f"An unexpected error {e} occurred when trying to translate \"{text_to_translate}\" "
                               f"from \"{language_to_translate_from}\" "
                               f"to \"{language_to_translate_to}\" "
                               f"using {translation_service} service")
                continue

            logger.info(f"Successfully translated \"{text_to_translate}\" to \"{translation}\" "
                        f"from \"{language_to_translate_from}\" "
                        f"to \"{language_to_translate_to}\" "
                        f"using {translation_service}")
            break
        else:
            logger.warning(f"Couldn't translate {text_to_translate} from {language_to_translate_from} "
                           f"to {language_to_translate_to} using any service")
            raise TranslatorError("All the specified translation services errored out")

        return translation



import asyncio
import gc
import logging
import time

from telegram import Update
from telegram.ext import ContextTypes, Application, MessageHandler, filters
from translators.server import TranslatorError

from suskabot.services.translator import TranslatorService

logger: logging.Logger = logging.getLogger(__name__)


class TranslatorHandlers:
    _application: Application
    _translator_service: TranslatorService

    def __init__(
            self,
            application: Application,
            translator_service: TranslatorService
    ) -> None:
        self._application = application
        self._translator_service = translator_service

        self._application.add_handlers(
            [
                MessageHandler(filters.TEXT & ~filters.COMMAND, self.translate, block=False),
            ]
        )

    async def translate(self, update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
        del context  # unused

        text_to_translate: str = update.effective_message.text

        try:
            translation = self._translator_service.translate(text_to_translate)
        except NotImplementedError:
            await update.effective_message.reply_text("Can't translate this language yet :(")
            logger.error(f"couldn't translate \"{text_to_translate}\", couldn't find language to translate to")
            return
        except TranslatorError as e:
            await update.effective_message.reply_text("Service error, please try again later")
            logger.error(f"{e} occured when trying to translate \"{text_to_translate}\"")
            return

        logger.info(f"Successfully sent the translation for \"{text_to_translate}\" "
                    f"to {update.effective_user.full_name}")

        await update.effective_message.reply_text(translation)



import logging

from telegram.ext import Application

from suskabot.controllers.telegram_api.handlers.app import AppHandlers
from suskabot.controllers.telegram_api.handlers.translator import TranslatorHandlers
from suskabot.controllers.telegram_api.handlers.youtube_downloader import YTDownloaderHandlers
from suskabot.services.translator import TranslatorService
from suskabot.services.youtube_downloader import YTDownloaderService


logger: logging.Logger = logging.getLogger(__name__)


class TGBot:
    _application: Application
    _translator_service: TranslatorService
    _yt_downloader_service: YTDownloaderService

    def __init__(
            self,
            application: Application,
            translator_service: TranslatorService,
            yt_downloader_service: YTDownloaderService,
    ) -> None:
        self._application = application
        self._translator_service = translator_service
        self._yt_downloader_service = yt_downloader_service

        AppHandlers(self._application)
        logger.debug("Successfully applied application handlers")

        YTDownloaderHandlers(self._application, self._yt_downloader_service)
        logger.debug("Successfully applied youtube downloader handlers")

        TranslatorHandlers(self._application, self._translator_service)
        logger.debug("Successfully applied translator handlers")

    def start(self) -> None:
        logger.debug("Starting TG bot application..")
        self._application.run_polling()
